#' sfsGrading
#'
#' Functions for grading exams at SfS
#'
#' @docType package
#' @name sfsGrading
#' @aliases sfsGrading-package
#' @import utils
#' @import stats
#' @import graphics
#' @importFrom grDevices dev.off pdf
#' @import ggplot2
#' @import knitr
#' @importFrom xtable xtable
#' @importFrom tools md5sum
"_PACKAGE"


#' Read edoz-file
#'
#' Read file that was exported from edoz.
#' 
#' @param file File that was exported from edoz (both german or english format is supported)
#' @return List with two elements:
#' \itemize{
#'   \item edozTable: dataframe with edoz export
#'   \item orig.names: original col names as found in edoz file
#' }
#' @export

readEdoz <- function(file) {
  ## DESC: Reads file that was exported from edoz
  ## INPUT: File of edoz file
  ## OUTPUT: List with two elements:
  ##  - edozTable: dataframe with edoz export
  ##  - orig.names: original col names as found in edoz file
  edozFile <- read.table(file,
                         header = TRUE,
                         sep = "\t",
                         colClasses = "character",
                         check.names = FALSE,
                         ## make sure that colnames are preserved
                         encoding = "utf-8",
                         quote = "\"") ## ow names like D'Adamo are with quote (')
  ## check for correct col names
  colNames.german <- c("Id", "Familienname", "Vorname", "Nummer",
                       "Sekretariat", "Fachrichtung", "Rep.",
                       "Note", "Notentendenz_+/-",
                       "Abbruch_oder_nicht_erschienen_*",
                       "Pruefsumme")
  colNames.english <- c("Id", "Last_Name", "First_Name", "Number",
                        "Administration_Office", "Direction",
                        "Rep.", "Grade", "Grade_Trend_+/-",
                        "dropout/no_show_*", "Checksum")
  
  colNames.current <- colnames(edozFile)
  
  colNames.internal <- c("ID", "familyName", "firstName", "number",
                         "office", "direction", "repeater", "grade",
                         "trend", "dropout", "checksum")      
                    
  if (length(colNames.current) == 11) {
    ## session examination 
    if (!(identical(colNames.current, colNames.german) |
          identical(colNames.current, colNames.english)) ) 
      stop("Column names in edoz file are not as expected.")
  } else {
    ## semester examination (does not contain column "trend" in eDoz
    if (!(identical(colNames.current, colNames.german[-9]) |
          identical(colNames.current, colNames.english[-9])) ) 
      stop("Column names in edoz file are not as expected.")
    colNames.internal <- colNames.internal[-9] ## remove position of "trend"
  }
  ## Overwrite column names with our internal version
  colnames(edozFile) <- colNames.internal 
  class(edozFile[,"grade"]) <- "numeric"
  
  list(edozTable = edozFile, orig.names = colNames.current)
}

#' Read Semester Performance File
#'
#' @param file csv-file with column number (student ID) and column gradeBonus which must be numeric between 0 and 0.25.
#' @param ... currently not used
#'
#' @return data frame with input columns
#' @export
#'
readSemesterPA <- function(file, ...){
  input <- read.csv(file = file, header = TRUE,
                    colClasses = c("character", NA), fill = TRUE,
                    stringsAsFactors = FALSE, na.strings = "-", ...)
  if(!all(colnames(input) == c("number", "gradeBonus")))
     stop("Column names of semester performance file must be number and gradeBonus")
  
  if(!is.numeric(input[,"gradeBonus"]))
    stop("Nonstandard values (characters?) in column gradeBonus")
  
  if(any(is.na(input[,"number"])))
    stop("Some student numbers are missing")
  
  ## Convert Legi-Nr to correct format
  input[,"number"] <- fixLegiNr(input[,"number"])

  ## drop students with missing gradeBonus
  input <- na.omit(input)
     
  if(min(input[,"gradeBonus"] < 0) | max(input[,"gradeBonus"] > 0.25))
     stop("gradeBonus must be between 0 and 0.25")
  
  if(any(duplicated(input[,"number"])))
    stop("duplicated student numbers found in semester performance file")
     
  input
}
  
#' Read MC answerfile
#' 
#' Read MC answerfile
#'
#' @param file csv-file with answers from MC-correction machine; needs to have 
#' exactly one column with name 'number' (=Legi-number; will be converted to lower case before checking)
#' 
#' @param ... passed on to read.csv()
#'
#' @return dataframe with student answers
#' @export
#'
readMC.answers <- function(file, ...) {
  input <-  read.csv(file = file, header = TRUE, fill = FALSE,
                     colClasses = "character",
                     stringsAsFactors = FALSE, ...)
  ## make sure that column "number" or similar is (uniquely) available
  number.ind <- which(tolower(colnames(input)) == "number")
  if(length(number.ind) != 1)
    stop("No or multiple number column(s) found")
  ## from now on the name is "number" (could have been "Number" before etc.)
  colnames(input)[number.ind] <- "number"
  
  if(any(duplicated(input[,"number"])))
    stop("Duplicated numbers found in MC.answers")
  
  input
}

#' Read Moodle points file
#' 
#' Read file exported from Moodle containing the points per question for each student. The correction itself
#' is done within Moodle and therefore only resulting points per questions are exported. A 
#' question can consist of several sub-questions, but only the aggregated number of points 
#' for the whole question will be exported.
#'
#' @param file Moodle export file; NA string is '-'
#' @param ... passed on to read.csv()
#'
#' @return
#' \itemize{
#'   \item moodleFile: dataframe with points per question and Legi-number in standard format
#'   \item max.points: vector containing the max number of points per question
#' }
#' @export
readMoodle <- function(file, ...){
  moodleFile <- read.csv(file = file,
                         header = TRUE,
                         colClasses = "character",
                         check.names = FALSE, na.strings = "-",
                         ## make sure that colnames are preserved
                         ...)
  
  ## Replace Moodle names with internal names using regexp magic
  cn <- colnames(moodleFile)
  
  ## Extract maximal number of points first
  cn.ex <- cn[grep("[FQ]\\.? [0-9]+ /.*", cn)]
  names.cn.ex <- gsub(pattern = "[FQ]\\.? ([0-9]+) /.*", 
                      replacement = "points_\\1", x = cn.ex)
  
  max.points <- chrPoints2numeric(gsub(pattern = "[FQ]\\.? [0-9]+ /(.*)", 
                           replacement = "\\1", x = cn.ex))
  names(max.points) <- names.cn.ex
  
  ## Sanity check with total points
  cn.total <- cn[grep("(Grade|Bewertung)/.*", cn)]
  ## (?:A|B) makes non-capturing group in regexp
  total.points <- chrPoints2numeric(gsub(pattern = "(?:Grade|Bewertung)/(.*)", 
                             replacement = "\\1", x = cn.total))
  
  if(sum(max.points) != total.points)
    stop("Exercise specific max points do not sum up to total max points")
  
  ## Replace naming scheme of exercises to internal convention
  cn <- gsub(pattern = "[FQ]\\.? ([0-9]+) /.*", replacement = "points_\\1", x = cn)
  cn <- gsub(pattern = "(Grade|Bewertung)/.*", replacement = "pointsTotal", x = cn)
  cn <- gsub(pattern = "Matriculation number|Matrikelnummer", replacement = "number", x = cn)
  
  colnames(moodleFile) <- cn
  
  ## Moodle contains lines like "Gesamtdurchschnitt", we drop them
  ## TODO: "number" is a global variable
  row.keep <- (moodleFile[,"number"] != "")
  moodleFile <- moodleFile[row.keep,]
  ## the following has issue while checking: number is considered global variable
  ## moodleFile <- subset(moodleFile, subset = (number != "")) 
  
  ## Transform Moodle Legi-Nr. to standard format
  moodleFile$number <- sub("(.{2})(.{3})(.{3})", "\\1-\\2-\\3", moodleFile$number)
  
  points.cols <- grep("points", colnames(moodleFile))
  moodleFile[points.cols] <- lapply(moodleFile[points.cols], chrPoints2numeric)
  
  list(moodleFile = moodleFile, max.points = max.points)
}

#' Read file with points from manual correction
#'
#' Read file with points from manual correction; this file has to be created manually. 
#' If it does not exist yet, call 'make' once. Together with the edoz-file, this will
#' create a skeleton-file that can be filled in manually.
#'
#' @param file csv-file with points from manual correction; needs the following column names
#' \itemize{
#'   \item number: Legi-number
#'   \item pointsTotal: Total points per student (sum of points per questions)
#'   \item points_1,..., points_n: One column with points per question
#' }
#' @param ... passed on to read.csv()
#'
#' @return data frame with points per student
#' @export
readManual <- function(file, ...){
  manualFile <- read.csv(file = file,
                         header = TRUE,
                         ##colClasses = "character",
                         ...)
  
  ## Add checks wrt consistency
  if(!"number" %in% colnames(manualFile))
    stop("Column 'number' in manual.pointsFile is missing!")
  
  if(!"pointsTotal" %in% colnames(manualFile))
    stop("Column 'pointsTotal' in manual.pointsFile is missing!")
  
  points.cols <- grep("points_", colnames(manualFile))
  
  if(length(points.cols) == 0)
    stop("No column starting with points_ found in manual.pointsFile!")
  
  if(any(rowSums(manualFile[,points.cols]) != manualFile[,"pointsTotal"]))
     stop("Some total points are not correct!")
     
  manualFile
}

#' Create file for manual correction
#'
#' @param output.file Filename of resulting skeleton file
#' @param edoz.file edoz-file from which legi-number is used
#' @param n.questions number of question for which the skelton file should contain a column
#' (with column names points_1, ..., points_n)
#' @param ... currently not passed on
#'
#' @return Writes a csv-file as skeleton for filling in the results of manual correction
#' @export
createManual <- function(output.file, edoz.file, n.questions, ...){
  edozTable <- readEdoz(edoz.file)$edozTable[,c("familyName", "firstName",
                                                "number")]
  edozTable[,c(paste0("points_", 1:n.questions), "pointsTotal")] <- NA
  write.table(edozTable, file = output.file, sep = ",", row.names = FALSE,
              na = "")
}


#' Replace legi number of student
#'
#' This is inteded for MC-exams where students were given temporary legi-numbers (this might happen
#' if answer sheets were prepared and some were lost); several checks are performed to avoid accidents.
#'
#' @param MC.answers data frame with answers of students (e.g. by readMC.answers())
#' @param edozList list as produced by readEdoz()
#' @param number.source vector with numbers that should be replaced
#' @param number.dest vector of same length with final numbers after replacement;
#' number.source[i] will be replaced by number.des[i]
#'
#' @return data farme MC.answers with updated numbers
#' @export
replaceNumber <- function(MC.answers, edozList, number.source, number.dest){
  
  edoz.numbers <- edozList$edozTable[,"number"]
  MC.answers.number <- MC.answers[,"number"]
  
  ## number.source and number.dest need to have same length
  if(length(number.source) != length(number.dest))
    stop("number.replace.source and number.replace.dest differ in length")
  
  ## make sure that number.source and number.dest do not contain duplicates
  if(any(duplicated(number.source)) | any(duplicated(number.dest)))
    stop("number.source or number.dest contain duplicated values")
  
  ## check whether number.source and number.dest are both of type character
  if(!(is.character(number.source) & is.character(number.dest)))
    stop("number.replace.source and number.replace.dest must be of type character")
    
  ## make sure that number.source is not in edoz
  if(any(number.source %in% edoz.numbers))
    stop("number.replace.source is an existing edoz number")
  
  ## make sure that number.source is not in MC.answers
  if(!all(number.source %in% MC.answers.number))
    stop("number.replace.source not found in MC.answers")
    
  ## make sure that number.dest is in setdiff(edoz, MC.answers)
  if(!all(number.dest %in% setdiff(edoz.numbers, MC.answers.number)))
    stop("number.replace.dest is not in setdiff(edoz, MC.answers)")

  pos <- match(number.source, table = MC.answers.number)
  
  MC.answers[,"number"] <- replace(MC.answers[,"number"], list = pos,
                                   values = number.dest)
  MC.answers
}

#' Read MC-key file with solutions
#'
#' @param file Text file with three cols with these names:
#' \itemize{
#' \item Exercise: Unique identifier for each exercise
#' \item CorrectAnswer: Correct value for each exercise. Use "-" to treat any or no answer as 
#' correct.
#' \item Index: Grouping of exercises (e.g. 1a, 1b, 1c all get Index=1; 2a, 2b, 2c get Index=2);
#' even there is no inherent structure of the excercises we RECOMMEND to group exercises in groups
#' of up to 8 subproblems. This makes reading the results after correction easier.
#' }
#' @param type Either \code{"TF"} for TRUE/FALSE questions or \code{"AB"} for questions where 
#' the correct options between e.g. a) - d) has to be found (single choice questions).
#' 
#' @param ... options passed on to read.csv
#'
#' @return data frame of the key; T/F is converted to a/b
#' @export
#'

readMC.key <- function(file, type, ...) {
  input <- read.csv(file, header = TRUE,
                    colClasses = "character",
                    stringsAsFactors = FALSE, ...)
  if(!all(c("Exercise", "CorrectAnswer", "Index") %in% colnames(input)))
    stop("MC.key does not contain column Exercise, CorrectAnswer and Index")
  inputString <- paste(tolower(input[,"CorrectAnswer"]), collapse = "")
  if(type == "TF"){ ## replace TF with AB
    ## check whether correct answers are indeed encoded as TF; no
    ## other characters are allowed
    if (grepl(pattern = "[^tf-]", inputString))
      stop("Wrong characters in MC.key!")
    ## Replace value with "ab"; we have to be careful because we allow
    ## character strings with length > 1 if multiple answers should be
    ## correct
    input[,"CorrectAnswer"] <- gsub(pattern = "t", replacement = "a",
                                    x = tolower(input[,"CorrectAnswer"]))
    input[,"CorrectAnswer"] <- gsub(pattern = "f", replacement = "b",
                                    x = tolower(input[,"CorrectAnswer"]))
  }else if(type == "AB"){
    ## check whether correct answers are indeed encoded as ABC..., no
    ## other characters are allowed
    ## TODO: Problem that AB is not converted to lower-case ???
    if (grepl(pattern = "[^a-z-]", inputString)) ## last - is for all answers correct
      stop("Wrong characters in MC.key!")
    ## do nothing
  }else{
    stop("Wrong type argument!")
  }
  
  input
}

#' Make sanity checks on MC.key and MC.answers
#'
#' @param MC.key data frame with solution
#' @param MC.answers data frame with student answers
#' @param MC.prefix string indicating the columns with ticked results
#'
#' @export
#'

check.answers.key <- function(MC.key, MC.answers, MC.prefix) {
  ## check for unique excercise labels
  ex.names <- MC.key[,"Exercise"]
  if(any(duplicated(ex.names))) 
    stop("Duplicated exercise labels found in MC key!")
  
  if(any(MC.key[,"CorrectAnswer"] == ""))
    stop("Empty string found in colum CorrectAnswer!")
  
  if(nrow(MC.key) == 0)
    stop("MC key is empty!")
  
  ## determine columns containing ticked answers
  tmp <- colnames(MC.answers)[grepl(pattern = MC.prefix,
                                    x = colnames(MC.answers))]
  ## remove prefix
  answersID  <- gsub(pattern = MC.prefix, replacement = "", x = tmp)
  
  ## answersID is potentially larger because of further information like
  ## exam type etc.
  stopifnot(all(ex.names %in% answersID))
  
  ## check if exercise labels in key and answer file are identical
  ##if (length(answersID) != length(ex.names))
  ##    ## setequal discards duplicated values, hence we need to compare
  ##    ## length too
  ##    stop("Not the same number of excercises in key and answers.")
  ##if (!setequal(ex.names, answersID))
  ##     stop("Exercise labels in key and answers don't match.")
}

#' Mark individual student answers as wrong or correct
#'
#' @param answers data frame with student answers
#' @param solutions data frame with correct solutions. Use "-" to treat any or no answer as 
#' correct.
#' @param MC.prefix string with prefix of colnames containing the ticked answers of the students
#'
#' @return List with two elements:
#' \itemize{
#' \item MC.corrections: data frame in the format of input answers but with entries TRUE/FALSE or NA (no answer by student)
#' \item MC.crit.cases: data frame in the format of MC.corrections where unexpected answers are collected
#' }
#' @export
#'

correctMC <- function(answers, solutions, MC.prefix){
  ## TESTED
  n.students  <- nrow(answers)
  n.questions <- nrow(solutions)
  
  corrections <- answers ## corrected answers
  
  ## Remove potentially available columns containing
  ## the original task name, otherwise these names will interfere with our
  ## output names
  ex.names    <- solutions[,"Exercise"]
  sel.drop    <- colnames(corrections) %in% ex.names
  corrections <- corrections[,!sel.drop]

  ## Fill exercise columns with missing values
  ## ex.cols <- grep(MC.prefix, colnames(answers))
  ex.cols  <- which(colnames(corrections) %in% paste0(MC.prefix, ex.names))
  corrections[,ex.cols] <- NA
  
  critical.cases <- corrections ## collection of unexpected answers
    
  for(i in 1:n.students){
    student.current <- answers[i,]
    
    for(j in 1:n.questions){ ## ordering according to MC.key
      question.ID <- solutions[j, "Exercise"]
      col.current <- paste0(MC.prefix, question.ID)
      
      answer.current <- tolower(student.current[,col.current])
      sol.current    <- tolower(solutions[j, "CorrectAnswer"])
      
      ## If mc.key is the special character "-" treat any or no answer as
      ## correct
      if(sol.current == "-"){
        corrections[i, col.current] <- TRUE
      }else{
        nchar.current <- nchar(answer.current)
        if(nchar.current > 1){ ## invalid input ## TODO: Triple check!!!
          ## Unexpected multiple selected answers
          ## Cases will be collected as "critical cases"
          critical.cases[i, col.current] <- answer.current
          
          ## If solution also has more options, we count every subset of
          ## the solution as correct (even for multiple answers)
          corrections[i, col.current] <- 
            if(nchar(sol.current) > 1){
              all(unlist(strsplit(answer.current, "")) %in% unlist(strsplit(sol.current, "")))
            }else{
              FALSE 
            }
        } else if(nchar.current == 0){ ## No answer
          corrections[i, col.current] <- NA
        } else if(nchar.current == 1){ ## Standard case: Single answer
          corrections[i, col.current] <-
            grepl(answer.current, sol.current)
          ## allows for multiple correct answers (e.g., if problems arise
          ## after exam)
        }else{
          stop("Something unexpected happens here!")
        }
      }
    }
  }
  colnames(corrections)[ex.cols] <- gsub(pattern = MC.prefix,
                                         replacement = "",
                                         x = colnames(corrections)[ex.cols])
  colnames(critical.cases) <- colnames(corrections)
  
  critical.cases.red <- critical.cases[,ex.cols]
  rownames(critical.cases.red) <- critical.cases[,"number"]
  ## Make sure that columns have same ordering as exercises in MC.key
  critical.cases.red <- critical.cases.red[,ex.names]
  
  ## Count number of ticks per student
  corrections[, "numberTicked"] <- apply(corrections[,ex.names], 1,
                                         function(x) sum(!is.na(x)))
  ## Remove prefix in corrections and critical cases data-frames
  
  ##colnames(corrections) <- gsub(pattern = MC.prefix, replacement = "",
  ##                              x = colnames(corrections))
  ##colnames(critical.cases) <- gsub(pattern = MC.prefix, replacement = "",
  ##                                 x = colnames(critical.cases))

  list(MC.corrections = corrections, MC.crit.cases = critical.cases.red)
}

#' Helper function for giving points
#'
#' @param x vector with TRUE/FALSE entries indicating correct/wrong solution
#' @param correct number of points for each correct solution (e.g. 1)
#' @param wrong number of points for each wrong solution (e.g. 0 or -0.5)
#' @param type 'SC' for single choice, 'SC-capped' for single choice capped at zero
#' points per exercise from below or 'kprim'; in 'kprim' there must be 4 answers per question. 
#' If you have all 4 correct, you get 2 points; if you have 3 out of 4 correct, you get 1 point.
#' Otherwise, you get 0 points. I.e., 'correct' and 'wrong' is not used in this option. NA is counted as wrong answer.
#'
#' @return Points (scalar) obtained
#'

points.ex <- function(x, correct, wrong, type){
  ## TESTED
  if(type == "SC"){
    sum(ifelse(x, correct, wrong), na.rm = TRUE)
  }else if(type == "kprim"){
    point.tmp <- sum(ifelse(x, 1, 0), na.rm = TRUE)
    max(point.tmp - 2, 0) ## kprim in functional form
  }else if(type == "SC-capped"){
    max(sum(ifelse(x, correct, wrong), na.rm = TRUE), 0)
  }else{
    stop("Type not supported!")
  }
}

#' Compute points obtained for each student
#'
#' @param corrections data frame with info on whether students ticked correct answer (TRUE/FALSE)
#' @param MC.key data frame with solutions and grouping info
#' @param correct number of points for each correct answer
#' @param wrong number of points for each wrong answer
#' @param type 'SC' for single choice, 'SC-capped' for single choice capped at zero
#' points per exercise from below or 'kprim'; in 'kprim' there must be 4 answers per question. 
#' If you have all 4 correct, you get 2 points; if you have 3 out of 4 correct, you get 1 point.
#' Otherwise, you get 0 points. I.e., 'correct' and 'wrong' is not used in this option. NA is counted as wrong answer.
#'
#' @export 
#' @return data frame with points per student (both for groups of excercises unter 'points_' and total under 'pointsTotal')
#'

givePoints <- function(corrections, MC.key,
                       correct = 1, wrong = 0, type = "SC"){
  ## TESTED
  ## Make sure that no column contains "points_"
  if(any(grepl("points_", colnames(corrections))))
    stop("Unallowed names points_ detected in colnames")
  
  ## Make sure that no column contains "pointsTotal"
  if(any(grepl("pointsTotal", colnames(corrections))))
    stop("Unallowed names pointsTotal detected in colnames")
  
  ind <- MC.key[, "Index"]  ## exercise structure
  ind.unique <- unique(ind) ## unique tasks
  
  ## Check for kprim:
  ## Each task needs to contain exactly 4 sub-tasks
  if(type == "kprim" & !isTRUE(all.equal(4, unique(table(ind)))))
    stop("Index column is not suitable for kprim exam")
    
  for(i in ind.unique){
    ## Extract all sub-tasks corresponding to current task
    current.ex <- MC.key[MC.key[,"Index"] == i, "Exercise"]
    
    ## Determine points for current task
    corrections[,paste0("points_", i)] <- 
      apply(corrections[,current.ex, drop = FALSE], 1, points.ex,
            correct = correct, wrong = wrong, type = type)
  }
  ## Sum up all points to get total points
  points.cols <- corrections[,grepl("points_", colnames(corrections))]
  corrections[,"pointsTotal"] <- rowSums(points.cols)
  corrections
}

#' Round value and prepare for pretty printing
#'
#' @param value raw value
#' @param digits number of digits to round to
#'
#' @return Character vector as returned from sprintf()
#' @export
#'

roundz <- function(value, digits){
  sprintf(paste("%1.", digits, "f", sep = ""), value)
}

##' Rounds to quarter grades
##'
##' @param x raw grade
##' @export
gradeRound <- function(x){
  ## TESTED
  pmax(1, pmin(6, floor(4 * x + 1/2) / 4))
    ## + 1 / 2 because of round-to-even issue
}

##' Linear grading scale, defined by points necessary for 4 and 6, resp.
##'
##' @param   points    vector of (total) points
##' @param   options   list of options; needs "pointsFor4" and "pointsFor6"
##'
##' @return  vector of raw grades
##' @export
gradesLinear <- function(points, options){
  ## TESTED
  6 - 2 * (options$pointsFor6 - points) /
    (options$pointsFor6 - options$pointsFor4)
}
##' Piecewise linear grading scale, defined by points necessary for 4 and 6,
##' resp., and the condition that 0 points correspond to grade 1
##'
##' @param   points    vector of (total) points
##' @param   options   list of options; needs "pointsFor4" and "pointsFor6"
##' @return  vector of raw grades
##' @export
gradesPiecewiseLinear <- function(points, options){
  ## TESTED
  grades <- numeric(length(points))
  failed <- points <= 0
  sufficient <- points >= options$pointsFor4 & points <= options$pointsFor6
  bestgrade <- points >= options$pointsFor6

  grades[bestgrade] <- 6
  grades[failed] <- 1

  b46 <- 2 / (options$pointsFor6 - options$pointsFor4)
  a46 <- 6 - b46 * options$pointsFor6
  grades[sufficient] <- a46 + b46 * points[sufficient]

  b14 <- 3 /options$pointsFor4
  a14 <- 1
  remainder <- !sufficient & !bestgrade & !failed
  grades[remainder] <- 1 + b14 * points[remainder]

  return(grades)
}

##' Piecewise linear grading scale, defined by points necessary for 4 and 6,
##' resp., and the condition that expected number of points through random guessing correspond to grade 1
##'
##' @param   points    vector of (total) points
##' @param   options   list of options; needs "pointsFor1" "pointsFor4" and "pointsFor6"
##' @return  vector of raw grades
##' @export
gradesPiecewiseLinearIntercept <- function(points, options)
{
  ## TESTED
  grades <- numeric(length(points))
  sufficient <- points >= options$pointsFor4
  grades[sufficient] <- 6 - 2*(options$pointsFor6 - points[sufficient])/
    (options$pointsFor6 - options$pointsFor4)
  above1 <- !sufficient & points >= options$pointsFor1
  grades[above1] <- 1 + 3 * (points[above1] - options$pointsFor1) / (options$pointsFor4 - options$pointsFor1)
  below1 <- !sufficient & !above1
  grades[below1] <- 1
  return(grades)
}

#' Compute grades per student
#'
#' @param pointsTotal data frame with points per student
#' @param gradeFn function for computing grade given points
#' @param settings further argument passed to gradeFn
#'
#' @return Matrix with two columns: rawGrade and (rounded) grade
#' @export
#'

applyGrading <- function(pointsTotal, gradeFn, settings){
  ## TESTED
  rawGrade <- rep(NA, length(pointsTotal))
  s <- is.na(pointsTotal)

  rawGrade[!s] <- gradeFn(pointsTotal[!s], option = settings)
  grade <- gradeRound(rawGrade)
  return(cbind(rawGrade = rawGrade, grade = grade))
}

#' Write grades to edoz-file
#'
#' Merge grade info into edoz-file and export.
#'
#' @param edozList list with edoz-infos as returned from readEdoz
#' @param results data frame with grades and further info
#' @param file filename of output file
#'
#' @return writes to file on disk
#' @export
#'

exportGrades <- function(edozList, results, file){
  edoz.out <- edozList$edozTable
  ## remove original empty grade column
  col.out <- !(colnames(edoz.out) == "grade")
  edoz.out <- edoz.out[,col.out]
  ## edoz.out <- subset(edoz.out, select = -grade) ## issue: grade counts as global var in check

  ## merge our grades to edoz.out
  points.cols <- colnames(results)[grepl("points_", colnames(results))]
  
  ## sel is the collection of variables that are needed for the overview PDF
  ## Some columns are only available for certain scenarios, e.g. semester 
  ## performance has rawGradeExam and gradeBonus
  sel <- intersect(c("number", points.cols, "pointsTotal", "rawGrade", 
                     "grade", "roundEffect", "rawGradeExam", "gradeBonus"), 
                   colnames(results))
  
  edoz.out <- merge(edoz.out, results[, sel], by = "number", all.x = TRUE)
  
  ## sort edoz.out according to family name and first name, ignoring
  ## lowercase / uppercase issues (because of family names like "von ...")
  edoz.out <- edoz.out[with(edoz.out, order(tolower(familyName),
                                            tolower(firstName))),]

  ## select original columns, including the corresponding ordering
  edoz.final <- edoz.out[, colnames(edozList$edozTable)]

  ## Perform checks
  
  ## Check if there are students which have a grade but are listed as 
  ## dropout
  stopifnot("dropout" %in% colnames(edoz.final))
  
  idxProblems <- !is.na(edoz.final[,"grade"]) & (edoz.final[,"dropout"] != "")
  if(any(idxProblems)){
    stop("Students marked as dropouts received a grade. Legi-Nr:", 
         paste(edoz.final[idxProblems,"number"], collapse = ","))
  }
  
  ## write file with original column names (by construction they have
  ## the same ordering) 
  write.table(edoz.final, file = sub(".txt", "-results.txt", file),
              quote = FALSE, sep = "\t", eol = "\r\n", na = "",
              row.names = FALSE, col.names = edozList$orig.names, 
              fileEncoding = "UTF-8")

  edoz.out ## output will be used for construction of results.pdf
}

#' Check written edoz-file
#'
#' Re-read written edoz-file and compare with grade infos; furthermore compute info on unusual student status (no-shows, etc.)
#'
#' @param file filename of edoz-file
#' @param results data frame with grading info per student
#'
#' @return List with four elements:
#' \itemize{
#' \item no.shows: Students who enrolled in edoz but did not show up (vector with Legi-numbers)
#' \item illegals: Students who did not enroll in edoz but showed up (vector with Legi-numbers)
#' \item nr.regulars: Nmb of students enrolled in edoz and who showed up
#' \item regulars.check: TRUE if all grades in re-read edoz-file agree with current data
#' }
#' @export
#'

checkResults <- function(file, results){
  resultsFile <- read.table(file,
                            header = TRUE,
                            sep = "\t",
                            colClasses = "character",
                            check.names = FALSE,
                            ## make sure that colnames are preserved
                            encoding = "utf-8",
                            quote = "\"")
  
  grade.col   <- which(colnames(resultsFile) %in% c("Note", "Grade"))
  number.col  <- which(colnames(resultsFile) %in% c("Nummer", "Number"))

  if(length(grade.col) == 0 | length(number.col) == 0)
    stop("No grade or number column found in edoz-export")
  if(length(grade.col) > 1 | length(number.col) > 1)
    stop("Multiple grade or number columns found in edoz-export")

  ## no shows: in eDoz but not in results
  no.shows <- setdiff(resultsFile[,number.col], results[,"number"])
  
  ## illegals: in results but not in eDoz
  illegals <- setdiff(results[,"number"], resultsFile[,number.col])
  
  ## regulars: both in eDoz and in results
  regulars <- intersect(resultsFile[,number.col], results[,"number"])
  
  if(length(regulars) == 0)
    stop("No student found which is both in eDoz and scan!")
  
  results.reg  <- results[results[,"number"] %in% regulars,]
  resultsFile.reg <- resultsFile[resultsFile[,number.col] %in% regulars,]
  
  match.vec <- match(results.reg[,"number"],
                     table = resultsFile.reg[,number.col])
  # 
  # ## match(c("B", "C", "A"), table = c("A", "B", "C"))
  # ## [1] 2 3 1

  regulars.check <- isTRUE(all.equal(results.reg[,"grade"], 
                                     as.numeric(resultsFile.reg[match.vec,
                                                                grade.col])))
  
  list(no.shows = no.shows, illegals = illegals, nr.regulars = length(regulars),
       regulars.check = regulars.check)
}

#' @exportS3Method NULL  
all.prop <- function(x){
  c(correct = sum(x,  na.rm = TRUE),
    wrong   = sum(!x, na.rm = TRUE),
    missing = sum(is.na(x))) / length(x)
}

distribPlot <- function(data, title, xlab, xmin, xmax, 
                        lowerLine = NULL, upperLine = NULL){
  distrib <- table(data)
  ##pdf(file, width = 3.5, height = 3)
  par(mar = c(3, 3, 3, 1), mgp = c(1.5, 0.6, 0))
  plot(names(distrib), as.numeric(distrib), xlim = c(xmin, xmax), 
       ylim = c(0, max(distrib)), xlab = xlab, ylab = "Frequency", 
       main = title, type = "h", cex = 0.66)
  if (!is.null(lowerLine))
    abline(v = lowerLine, col = 3, lty = 2)
  if (!is.null(upperLine))
    abline(v = upperLine, col = 4, lty = 2)
  ##dev.off()
  cat(sprintf("\n\\vspace{-3em}\n\n\\begin{center}Mean: %s; Median: %s; SD: %s \\end{center}\n",
              roundz(mean(data, na.rm = TRUE), 2),
              roundz(median(data,na.rm = TRUE), 2),
              roundz(sd(data, na.rm = TRUE), 2)))
}


## Generates an annotated distribution plot
distribPlot_old <- function(data, title, xlab, file, xmin, xmax, 
                            lowerLine = NULL, upperLine = NULL){
  distrib <- table(data)
  pdf(file, width = 3.5, height = 3)
  par(mar = c(3, 3, 3, 1), mgp = c(1.5, 0.6, 0))
  plot(names(distrib), as.numeric(distrib), xlim = c(xmin, xmax), 
       ylim = c(0, max(distrib)), xlab = xlab, ylab = "Frequency", 
       main = title, type = "h", cex = 0.66)
  if (!is.null(lowerLine))
    abline(v = lowerLine, col = 3)
  if (!is.null(upperLine))
    abline(v = upperLine, col = 4)
  dev.off()
  cat(sprintf("\\begin{minipage}{0.45\\textwidth}\n\\includegraphics[width=\\textwidth]{%s}\n{\\small \\centering Mean: %s; Median: %s; SD: %s }\n\\end{minipage}\\hspace{10mm}\n", file, roundz(mean(data, na.rm = TRUE), 2), roundz(median(data,na.rm = TRUE), 2), roundz(sd(data, na.rm = TRUE), 2)))
}

## Testing functions

#' Compare two edoz-files wrt. number and grade
#'
#' @param file1 edoz-file 1
#' @param file2 edoz-file 2
#'
#' @return TRUE if both files agree (after sorting) wrt number and grade
#' @export
#'

compareEdozTest <- function(file1, file2){
  d1 <- readEdoz(file1)$edozTable
  d2 <- readEdoz(file2)$edozTable
  
  sel <- c("number", "grade")
  d1Sorted <- d1[order(d1$number), sel]
  d2Sorted <- d2[order(d2$number), sel]
  rownames(d1Sorted) <- rownames(d2Sorted) <- NULL
  
  isTRUE(all.equal(d1Sorted, d2Sorted))
}

## Misc ####

truncate_string <- function(x, len = 20L, end = TRUE) {
  if(end){
    x[nchar(x) > len] <- paste0(substr(x[nchar(x) > len], 1L, len - 3L), "...")
  }else{
    x[nchar(x) > len] <- paste0("...", substr(x[nchar(x) > len], nchar(x) - len + 4, nchar(x)))
  }
  x
}


getOverview <- function(x){
  round(c(mean(x, na.rm = TRUE), 
          median(x, na.rm = TRUE), 
          min(x, na.rm = TRUE),
          quantile(x, probs = 0.25, na.rm = TRUE), 
          quantile(x, probs = 0.75, na.rm = TRUE),
          max(x, na.rm = TRUE)), 2)
}

printSmallTable <- function(condition, x, pre = "\\\\[0.5em] ", 
                            include.rownames = FALSE, 
                            include.colnames = FALSE, 
                            hline.after = NULL, size = "small", 
                            floating = FALSE, latex.environments = NULL, ...){
  if(condition){
    paste0(pre, print(xtable::xtable(x), 
                      include.rownames = include.rownames, 
                      include.colnames = include.colnames, 
                      hline.after = hline.after, size = size, 
                      floating = floating, 
                      latex.environments = latex.environments, ...))
  } else {
    "None"
  }
}

chrPoints2numeric <- function(x) {
  ## x: character vector with entries like "1,00" (comma as decimal string) and phps other entries
  ## 
  ## if x is character with at least 1 element: 
  ## replace all "," by "." and then convert to numeric
  
  if (mode(x) == "character" & length(x) > 0) {
    res <- gsub(pattern = ",", replacement = "\\.", x=x)
  }
  as.numeric(res)
}

cleanPath <- function(path){
  ## See https://stackoverflow.com/questions/45619522/preventing-duplicate-slashes-in-file-paths
  sep <- .Platform$file.sep
  result <- gsub(paste0(sep,"{2,}"), sep, path, fixed = FALSE, perl = TRUE)
  return(result)
}

fixLegiNr <- function(x){
  ## Convert Legi-Nr. to standard format
  ## Check if in edoz format
  ok <- grepl(pattern = "^\\d{2}-\\d{3}-\\d{3}$", x = x)
  
  if(!all(ok)){
    ## check if the number is in 8-digit (e.g. moodle) format
    if(!all(grepl(pattern = "^\\d{8}$", x[!ok])))
      stop("Some Legi-numbers are in unknown format.")
    else{
      ## transform 8-digit format to edoz format
      x[!ok] <- sub("(.{2})(.{3})(.{3})", "\\1-\\2-\\3", x[!ok])
    }
  }
  x  
}
